##################################################################################################################################
#                                                   BivarEGEE                                                                    #
##################################################################################################################################
##           Author: Shuai Wang tutuwang@bu.edu                                                                                  #
##          Version: beta (intended for family data only)                                                                        #
##  Date of Release: 02/28/2015                                                                                                  #
##################################################################################################################################
##  Description of the arguments                                                                                                ##
##  formula1  : the formula of the continuous phenotype without the genotype term. e.g. bmi~age                                 ##
##  formula2  : the formula of the binary phenotype without the genotype term, e.g. t2d~age+sex                                 ##
##  g         : the column name of the genotype in the dataframe data, must be quoted in quotation marks, e.g. "g1"             ##
##  id        : the column name of the id identifier in the dataframe data, must be quoted in quotation marks, e.g. "ID"        ##
##  famid     : the column name of the famid identifier in the dataframe data, must be quoted in quotation marks, e.g. "FAMID"  ##
##  data      : the name of the dataframe.                                                                                      ##
##  fullkins  : the kinship matrix                                                                                              ##
##  tol       : the convergence criteria                                                                                        ##
##  restricted: by default restricted=F, the unrestricted estimation is implemented                                             ##
##################################################################################################################################

BivarEGEE=function(formula1, formula2, g, id, famid, data, fullkins, tol=10^(-8), restricted=FALSE ){
fullkins=as.matrix(fullkins)
library(gdata)
phen1=paste(formula1)[2]
phen2=paste(formula2)[2]
cov1=paste(formula1)[3]
cov2=paste(formula2)[3]
if ((grepl("-1", deparse(formula1))==T)  |(grepl("- 1", deparse(formula1))==T)){
   
    cov1=sapply(strsplit(cov1,"+", fixed=T), function(x){trim(unlist(strsplit(x, "-", fixed=T)))})
    cov1=cov1[cov1!="1"]
} else {
    cov1=unlist(strsplit(cov1, "+", fixed=T))
    cov1=sapply(cov1, trim)
}

cov1=cov1[nchar(cov1)>0]

if ((grepl("-1", deparse(formula2))==T)  |(grepl("- 1", deparse(formula2))==T)){
   
    cov2=sapply(strsplit(cov2,"+", fixed=T), function(x){trim(unlist(strsplit(x, "-", fixed=T)))})
    cov2=cov2[cov2!="1"]
} else {
    cov2=unlist(strsplit(cov2, "+", fixed=T))
    cov2=sapply(cov2, trim)
}

cov2=cov2[nchar(cov2)>0]

data=data[, unique(c(id, famid, phen1, phen2, cov1, cov2, g))]
data=data[apply(data, 1, function(x){all(!is.na(x))}),]
kmat=as.matrix(fullkins)
#kmat2=kmat[as.character(data[,id]), as.character(data[,id])]
X1_ALL=cbind(rep(1, nrow(data)), data[, cov1])
X2_ALL=cbind(rep(1, nrow(data)), data[, cov2])
lm1=lm(formula1, data=data)
beta1<-summary(lm1)$coefficients[,1]
beta1=matrix(beta1, ncol=1)
beta2<-as.vector(coef(glm(formula2,data=data, family="binomial")))
beta2=matrix(beta2, ncol=1)
sigma1sq<-1
thai=1
#sigma12<-0.5
sigmaesq<-(summary(lm1)$sigma)^2
r=cor(data[,phen1], data[,phen2])
r.abs=abs(r)
r.sign=sign(r)
sqerror=1
iter=0

while (sqerror>=tol){
  
  Ustar=matrix(0, ncol=(ncol(X1_ALL)+ncol(X2_ALL) + 4), nrow=(ncol(X1_ALL)+ncol(X2_ALL) + 4 ))
  U=matrix(0, ncol=1, nrow=(ncol(X1_ALL)+ncol(X2_ALL) + 4))
 
  for (i in  unique(data[, famid])) {
    
    v.famid <- data[,famid]==i
    ni=sum(v.famid)
    id.tmp=as.character(data[v.famid, id])
  
    Sigmakin_i = kmat[id.tmp, id.tmp]
    r.mat=ifelse(Sigmakin_i==0, 0, r.sign*(r.abs)^(1/2/Sigmakin_i))
    
    X1=X1_ALL[v.famid,]
    X1=as.matrix(X1)
    X2=X2_ALL[v.famid,]
    X2=as.matrix(X2)

    if (ncol(X1_ALL)==1){
      X1=matrix(X1, ncol=1)
    }

    if (ncol(X2_ALL)==1){
      X2=matrix(X2, ncol=1)
    }
    
          
    mu1=X1 %*% beta1
    pred2=X2%*%beta2
    
    exp.pred2 = exp(pred2)
    mu2=exp.pred2/(1+exp.pred2)
    
    phen1_i=data[v.famid, phen1]
    phen2_i=data[v.famid, phen2]
    
    si = as.vector(matrix(c(phen1_i-mu1, phen2_i-mu2), ncol=1) %*% matrix(c(phen1_i-mu1, phen2_i-mu2), nrow=1))
    
    D=matrix(0, nrow=2*ni, ncol=(ncol(X1_ALL)+ncol(X2_ALL)))
    D[1:ni, (1:ncol(X1_ALL))] = X1
    D[(ni+1):(2*ni), ((ncol(X1_ALL)+1):ncol(D))] = as.numeric(exp.pred2/(1+exp.pred2)^2) * X2  
    
   
    Var_cont = sigmaesq * diag(rep(1, ni)) + 2*sigma1sq * Sigmakin_i
    R.w=2*Sigmakin_i*thai
    diag(R.w)=1
    Var_bin=diag(sqrt(as.vector(mu2*(1-mu2))))%*%R.w%*%diag(sqrt(as.vector(mu2*(1-mu2))))
    Cov = diag(sqrt(diag(Var_cont))) %*% r.mat %*% diag(sqrt(diag(Var_bin)))
    Vi=rbind(cbind(Var_cont, Cov), cbind(t(Cov), Var_bin))
         
  
   solve.Vi = solve(Vi)
    sigma_i = as.vector(Vi)
    Fi = matrix(0, ncol=4, nrow=ni^2*4)
    Vi_sigma1sq = matrix(0, ncol=(ni*2), nrow=(ni*2))
    Vi_sigma1sq[1:ni,] = cbind(2*Sigmakin_i, diag(1/2/sqrt(sigmaesq+sigma1sq), ni)*r.mat%*%diag(sqrt(diag(Var_bin))))
    Vi_sigma1sq[(1+ni):(2*ni),(1:ni)] =t(Vi_sigma1sq[(1:ni), (1+ni):(2*ni)])
    
    
    Fi[,1]=as.vector(-solve.Vi %*% Vi_sigma1sq %*% solve.Vi)
    R.w_thai=2*Sigmakin_i
    diag(R.w_thai)=0
    Var_bin_thai=diag(sqrt(as.vector(mu2*(1-mu2))))%*%R.w_thai%*%diag(sqrt(as.vector(mu2*(1-mu2))))

    
    Vi_thai=rbind(matrix(0, nrow=ni,ncol=(2*ni)), cbind(matrix(0, nrow=ni,ncol=ni), Var_bin_thai))
   
    Fi[,2]=as.vector(-solve.Vi %*% Vi_thai %*% solve.Vi)
    
    Vi_sigmaesq=matrix(0, nrow=(ni*2), ncol=(ni*2))
    Vi_sigmaesq[(1:ni), (1:ni)]=diag(rep(1, ni))
    Vi_sigmaesq[(1:ni), ((1+ni):(ni*2))]=diag(1/2/sqrt(diag(Var_cont))) %*% r.mat %*% diag(sqrt(diag(Var_bin)))
    Vi_sigmaesq[((1+ni):(ni*2)), (1:ni)]=t(Vi_sigmaesq[(1:ni), ((1+ni):(ni*2))])
    Fi[,3]=as.vector(-solve.Vi %*% Vi_sigmaesq %*% solve.Vi)
    Vi_r=matrix(0, nrow=(ni*2), ncol=(ni*2))

    Vi_r[1:ni, ((ni+1):(ni*2))]=diag(sqrt(diag(Var_cont)))%*%
         ifelse(Sigmakin_i == 0,0, r.sign*1/2/Sigmakin_i*(r.abs)^(1/2/Sigmakin_i-1)) %*%diag(sqrt(diag(Var_bin)))
     Vi_r[((ni+1):(ni*2)), 1:ni]=t(Vi_r[1:ni, ((ni+1):(ni*2))])

    Fi[,4]=as.vector(-solve.Vi %*% Vi_r %*% solve.Vi)
    mu2.vec=as.vector(mu2)
    Vi_alpha=cbind(as.vector(Vi_sigma1sq), as.vector(Vi_thai),  as.vector(Vi_sigmaesq), as.vector(Vi_r))

    Vi_beta=cbind(matrix(0, nrow=(ni^2*4), ncol=ncol(X1_ALL)), sapply(1:ncol(X2_ALL), function(x){
Vi_bin_beta=diag(1/2/sqrt(mu2.vec*(1-mu2.vec))*(1-2*mu2.vec)*(1-mu2.vec)*mu2.vec*X2[,x])%*%R.w%*%diag(sqrt(mu2.vec*(1-mu2.vec)))
+diag(sqrt(mu2.vec*(1-mu2.vec)))%*%R.w%*%diag(1/2/sqrt(mu2.vec*(1-mu2.vec))*(1-2*mu2.vec)*(1-mu2.vec)*mu2.vec*X2[,x]);
Cov_beta=diag(sqrt(diag(Var_cont))) %*% r.mat %*% diag(1/2/sqrt(diag(Var_bin))*(1-2*mu2.vec)*(1-mu2.vec)*mu2.vec*X2[,x]);
return(as.vector(rbind(cbind(matrix(0, ni, ni), Cov_beta), cbind(t(Cov_beta), Vi_bin_beta))))
  }))

          
    Utemp <- t(D) %*% solve.Vi
    Ustar_entry = cbind(rbind(Utemp %*% D, t(Fi) %*% Vi_beta), 
                        rbind(matrix( 0, nrow = ncol(X1_ALL)+ncol(X2_ALL), ncol=4),t(Fi) %*% Vi_alpha) )
    U_entry =rbind(Utemp %*% c(phen1_i-mu1, phen2_i-mu2), t(Fi) %*% (si-sigma_i))
    
    Ustar <- Ustar + Ustar_entry
    U <- U + U_entry
  } 

  
  par.old=matrix(c(beta1, beta2, sigma1sq, thai, sigmaesq, r) , ncol=1)
  par.new=par.old+solve(Ustar) %*% U        
  sqerror=sum((par.old-par.new)^2)
  beta1=par.new[1:ncol(X1_ALL)]
  beta1=matrix(beta1, ncol=1)
  beta2=par.new[(ncol(X1_ALL)+1):(ncol(X1_ALL)+ncol(X2_ALL))]
  beta2=matrix(beta2, ncol=1)
  if (par.new[(ncol(X1_ALL)+ncol(X2_ALL)+1)]>=0){
      sigma1sq=par.new[(ncol(X1_ALL)+ncol(X2_ALL)+1)]
  }
  if (par.new[(ncol(X1_ALL)+ncol(X2_ALL)+2)]>=0){
    thai=par.new[(ncol(X1_ALL)+ncol(X2_ALL)+2)]
  }  
  #sigma12=par.new[(ncol(X1_ALL)+ncol(X2_ALL)+3)]
  if (par.new[(ncol(X1_ALL)+ncol(X2_ALL)+3)]>=0){
      sigmaesq=par.new[(ncol(X1_ALL)+ncol(X2_ALL)+3)]
  }
  if (sign(par.new[(ncol(X1_ALL)+ncol(X2_ALL)+4)])==r.sign){
  r=par.new[(ncol(X1_ALL)+ncol(X2_ALL)+4)]
  }
  r.abs=abs(r)
  r.sign=sign(r)
  iter=iter+1
 
}    

X1_ALL=cbind(X1_ALL, data[,g])
X2_ALL=cbind(X2_ALL, data[,g])


beta1=c(beta1,0)
beta2=c(beta2,0)
sandwich_I=matrix(0, ncol=(ncol(X1_ALL)+ncol(X2_ALL)), nrow=(ncol(X1_ALL)+ncol(X2_ALL)))
U=matrix(0, ncol=1, nrow=(ncol(X1_ALL)+ncol(X2_ALL)))
Ustar=matrix(0, ncol=(ncol(X1_ALL)+ncol(X2_ALL)), nrow=(ncol(X1_ALL)+ncol(X2_ALL)))


for(i in 1:N){
   v.famid <- data[,famid]==i
    ni=sum(v.famid)
    id.tmp=as.character(data[v.famid, id])
  
    Sigmakin_i = kmat[id.tmp, id.tmp]
    r.mat=ifelse(Sigmakin_i==0, 0, r.sign*(r.abs)^(1/2/Sigmakin_i))
    
    X1=X1_ALL[v.famid,]
    X1=as.matrix(X1)
    X2=X2_ALL[v.famid,]

    if (ncol(X2_ALL)==1){
      X2=matrix(X2, ncol=1)
    }
    X2=as.matrix(X2)
       
    mu1=X1 %*% beta1
     pred2=X2%*%beta2
    
    exp.pred2 = exp(pred2)
    mu2=exp.pred2/(1+exp.pred2)
    
    phen1_i=data[v.famid, phen1]
    phen2_i=data[v.famid, phen2]
    
    si = as.vector(matrix(c(phen1_i-mu1, phen2_i-mu2), ncol=1) %*% matrix(c(phen1_i-mu1, phen2_i-mu2), nrow=1))
    
    D=matrix(0, nrow=2*ni, ncol=(ncol(X1_ALL)+ncol(X2_ALL)))
    D[1:ni, (1:ncol(X1_ALL))] = X1
    D[(ni+1):(2*ni), ((ncol(X1_ALL)+1):ncol(D))] = as.numeric(exp.pred2/(1+exp.pred2)^2) * X2  
    
   
    Var_cont = sigmaesq * diag(rep(1, ni)) + 2*sigma1sq * Sigmakin_i
    R.w=2*Sigmakin_i*thai
    diag(R.w)=1
    Var_bin=diag(sqrt(as.vector(mu2*(1-mu2))))%*%R.w%*%diag(sqrt(as.vector(mu2*(1-mu2))))
    Cov = diag(sqrt(diag(Var_cont))) %*% r.mat %*% diag(sqrt(diag(Var_bin)))
    Vi=rbind(cbind(Var_cont, Cov), cbind(t(Cov), Var_bin))
         
  
   solve.Vi = solve(Vi)
    sigma_i = as.vector(Vi)
sandwich_I=sandwich_I+t(D)%*%solve.Vi%*%matrix(c(phen1_i-mu1, phen2_i-mu2), ncol=1) %*% matrix(c(phen1_i-mu1, phen2_i-mu2), nrow=1)%*%solve.Vi%*%D
         Ustar.fac=t(D)%*%solve.Vi

    U.entry=Ustar.fac%*%matrix(c(phen1_i-mu1, phen2_i-mu2), ncol=1)
    U=U+U.entry
   Ustar=Ustar+Ustar.fac%*%D


} 


U_s=U[c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]
I_s=sandwich_I[c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL)), c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]
I_zx=sandwich_I[c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL)), -c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]
I_xx=sandwich_I[-c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL)), -c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]
I_s.adj=I_s-I_zx%*%solve(I_xx)%*%t(I_zx)

score_stats=matrix(U_s, nrow=1)%*%solve(I_s.adj)%*%matrix(U_s, ncol=1)

U1=U[-c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]
U=c(U1,U_s)

I_ideal=Ustar
C=cbind(-I_ideal[c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL)),-c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]%*%solve(I_ideal[-c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL)),-c(ncol(X1_ALL), ncol(X1_ALL)+ncol(X2_ALL))]),
    diag(1, 2))

sandwich_I=sandwich_I[c(1:(ncol(X1_ALL)-1), (ncol(X1_ALL)+1): (ncol(X1_ALL)+ncol(X2_ALL)-1), ncol(X1_ALL), (ncol(X1_ALL)+ncol(X2_ALL))), c(1:(ncol(X1_ALL)-1), (ncol(X1_ALL)+1): (ncol(X1_ALL)+ncol(X2_ALL)-1), ncol(X1_ALL), (ncol(X1_ALL)+ncol(X2_ALL)))]
robust_score=t(C%*%U)%*%solve(C%*%sandwich_I%*%t(C))%*%C%*%U

p.score=pchisq(robust_score, df=2, lower.tail=F)
p.wang=pchisq(score_stats, df=2, lower.tail=F)

if (restricted==FALSE){
    res.out=list(beta1=beta1, beta2=beta2, sigam1sq=sigma1sq,  r=r, thai=thai, iter=iter, statistic=robust_score, p=p.score)
} else if (restricted==TRUE){
    res.out=list(beta1=beta1, beta2=beta2, sigam1sq=sigma1sq,  r=r, thai=thai, iter=iter, statistic=score_stats, p=p.wang)
} else {
    stop ("please provide either a TRUE or FALSE value for the argument restricted!")        
}      
return(res.out)
}




